/* GKrellM HLT - Hot Little Therm Monitor
 * Copyright (C) 2000 Greg Hankins
 *
 * $Id: gkrellmhlt.c,v 1.13 2000/09/16 21:50:08 ghankins Exp $
 * 
 * Author: Greg Hankins <gregh@twougys.org>
 * Latest versions can be found at: http://www.twoguys.org/~gregh/software/
 * 
 * This program is free software which I release under the GNU General Public
 * License. You may redistribute and/or modify this program under the terms
 * of that license as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * To get a copy of the GNU General Puplic License, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*#define DEBUG*/

#include <string.h>
#include <gkrellm/gkrellm.h>
#define PLUGIN_CONFIG_KEYWORD "gkrellmhlt"
#define HLT_SAMPLE_DELAY 120

char *therm_main(int therm_argc, char **therm_argv);
char **gettherment(char *name);
char *searchent(char **te, char *name);
int  anytoi(char *p);
int  openline(char **te);
int  readsensor(char *ifmt, int fd, int ad, double *r);

static Panel *hlt_panel;
static Decal *hlt_decal;
static gint  hlt_style_id;
static gchar *hlt_temp;      /* temperature string                   */
static gchar hlt_text[21];   /* the actual text string that is drawn */
char         str[11];

/* thermometer entry */
static struct therm_entry {
    char   *ifmt;     /* sensor input format   */
    char   *conv;     /* type of conversion    */
    char   *therm;    /* name of therm to read */
    char   *format;   /* output format string  */
    double tof;       /* temperature offset    */
    int    ns;        /* number of samples     */
    int    ad;        /* address byte          */
    char   **te;      /* therm entry           */
} my_therm;

/* 
 * Read and convert temperature from the therm.
 */
void read_therm() {

    int i;
    int d;
    double t;
    double ts;
    double *ta;
    double tof = 0;

    ta = (double *)malloc(my_therm.ns * sizeof(double));
    if ((d = openline(my_therm.te)) == -1) {
        strncpy(str, "ERR2", 4);
        free(ta);
        return;
    }

    ts = 0;
    for (i = my_therm.ns; i-- > 0; ) {
        if (readsensor(my_therm.ifmt, d, my_therm.ad, &t) < 0) {
            close(d);
            strncpy(str, "ERR3", 4);
            free(ta);
            return;
        }
        t += tof;
        ta[i] = t;
        ts += t;
    }

    switch (my_therm.conv[0]) {
        case 'f':
            ts = (ts * 9) / 5 + 32;
            break;
        case 'c':
            ts = ((ts - 32) * 5) / 9;
            break;
        default:
            break;
    }
    snprintf(str, 10, "%2.2f%c", ts, toupper(my_therm.conv[0]));
#ifdef DEBUG
    fprintf(stderr, "%s\n", str);
#endif

    free(ta);
    close(d);
    return;
}

/*
 * I don't really know what this does, but it seems to be needed.
 */
static gint panel_expose_event(GtkWidget *widget, GdkEventExpose *ev) {

    if (widget == hlt_panel->drawing_area) {
        gdk_draw_pixmap(widget->window, 
                        widget->style->fg_gc[GTK_WIDGET_STATE(widget)], 
                        hlt_panel->pixmap, ev->area.x, ev->area.y, ev->area.x, 
                        ev->area.y, ev->area.width, ev->area.height);
    }
    return FALSE;
}


static void update_plugin() {
    
    static gint  w;                   /* string width                         */
    TextStyle    *ts;                 /* TextStyle of the string              */
    static gint  delay = 0;           /* delay counter                        */
    static gchar *hlt_name = "HLT";   /* name string                          */

    /* only sample every once in a while (120 "cycles" by default) */
    if (delay == 0) {
        if (my_therm.te == NULL) {
            snprintf(hlt_text, 20, "%s %s", hlt_name, hlt_temp);
        }
        else { 
            read_therm();
            snprintf(hlt_text, 20, "%s %s", hlt_name, str);
        }

        delay = HLT_SAMPLE_DELAY;
    }
    else {
        delay--;
    }

    /* get current settings */
    ts = gkrellm_meter_textstyle(hlt_style_id);
    w = gdk_string_width(ts->font, hlt_text);

    /* center text and set font */
    hlt_decal->x_off = (hlt_decal->w - w) / 2;
    if (hlt_decal->x_off < 0) {
        hlt_decal->x_off = 0;
    }
    hlt_decal->text_style.font = ts->font;

    /* draw */
    gkrellm_draw_text_decal(hlt_panel, hlt_decal, hlt_text, -1);
    gkrellm_draw_layers(hlt_panel);
}

/*
 * Create the HLT plugin.
 */
static void create_plugin(GtkWidget *vbox, gint first_create) {

    Style     *hlt_style;   /*     */
    TextStyle *ts;          /*     */
    TextStyle *ts_alt;      /*     */
    char      *p;           /* tmp */

    if (first_create) {
        hlt_panel = gkrellm_panel_new0();
    }
    else {
        gkrellm_destroy_decal_list(hlt_panel);
    }
    hlt_style = gkrellm_meter_style(hlt_style_id);
    hlt_style->label_position = LABEL_CENTER;

    /* Each Style has two text styles.  The theme designer has picked the
       colors and font sizes, presumably based on knowledge of what you draw
       on your panel.  You just do the drawing.  You probably could assume
       the ts font is larger than the ts_alt font, but again you can be
       overridden by the theme designer.                                      */
    ts = gkrellm_meter_textstyle(hlt_style_id);
    ts_alt = gkrellm_meter_alt_textstyle(hlt_style_id);
    hlt_panel->textstyle = ts;   /* would be used for a panel label */

    /* Create a text decal, make it the full panel width (minus the margins).
       Position at top border and left margin of the style.                   */
    hlt_decal = gkrellm_create_decal_text(hlt_panel, "Ay", ts, hlt_style, -1, 
                                          -1, -1);

    /* Configure the panel to hold the above created decals. */
    gkrellm_configure_panel(hlt_panel, NULL, hlt_style);
    gkrellm_create_panel(vbox, hlt_panel, gkrellm_bg_meter_image(hlt_style_id));
    gkrellm_monitor_height_adjust(hlt_panel->h);

    if (first_create) {
        gtk_signal_connect(GTK_OBJECT(hlt_panel->drawing_area), "expose_event",
                          (GtkSignalFunc)panel_expose_event, NULL);
    }

    /* setup therm defaults */
    my_therm.ifmt = "8f";
    my_therm.conv = "n";
    my_therm.format = "%f\n";
    my_therm.tof = 0;
    my_therm.ns = 1;
    my_therm.ad = 0;
    my_therm.therm = "misfits";

    /* get the therm entry */
    if (!(my_therm.te = gettherment(my_therm.therm))) {
#ifdef DEBUG
        fprintf(stderr, "unknown device: %s\n", my_therm.therm);
#endif
        hlt_temp = "ERR1";
        my_therm.te = NULL;
    }
    else {
        if ((p = searchent(my_therm.te, "is="))) my_therm.ifmt = p;
        if ((p = searchent(my_therm.te, "ct="))) my_therm.conv = p;
        if ((p = searchent(my_therm.te, "fs="))) my_therm.format = p;
        if ((p = searchent(my_therm.te, "to#"))) my_therm.tof = atof(p);
        if ((p = searchent(my_therm.te, "ns#"))) {
            my_therm.ns = anytoi(p);
            if (my_therm.ns < 1) my_therm.ns = 1;
            else if (my_therm.ns > 1000) my_therm.ns = 1000;
        }
        if ((p = searchent(my_therm.te, "ad#"))) my_therm.ad = anytoi(p);
#ifdef DEBUG
        fprintf(stderr, "got device: %s\n", my_therm.therm);
#endif
    }
}

static Monitor hlt_monitor = {
    "HLT",           /* Name, for config tab                     */
    0,               /* Id, 0 if a plugin                        */
    create_plugin,   /* The create function                      */
    update_plugin,   /* The update function                      */
    NULL,            /* The config tab create function           */
    NULL,            /* Apply the config function                */
    NULL,            /* Save user config                         */
    NULL,            /* Load user config                         */
    NULL,            /* config keyword                           */
    NULL,            /* Undefined 2                              */
    NULL,            /* Undefined 1                              */
    NULL,            /* private                                  */
    MON_UPTIME,      /* Insert plugin before this monitor        */ 
    NULL,            /* Handle if a plugin, filled in by GKrellM */
    NULL             /* path if a plugin, filled in by GKrellM   */
};

/*
 * Initialize the HLT plugin.
 */
Monitor *init_plugin() {
    hlt_style_id = gkrellm_add_meter_style(&hlt_monitor, DEFAULT_STYLE_ID);
    return &hlt_monitor;
}
