/*
*	$Id: therm.c,v 1.6 2000/09/11 04:41:49 ghankins Exp $
*
*	Read the temperature from a thermometer device.
*	Sends an address byte to device and reads a temperature byte back.
*
*	Uses a configuration file (RCFILE) reminiscent of termcap, printcap, ...
*	The field types are:
*		dv=   device file                                      [no default]
*		pa=   parity, one of 'e', 'o', 'm', 's'                [line default]
*		br#   baud rate                                        [line default]
*		ad#   address byte to send                             ['\n']
*		ns#   number of samples to average (for noisy sensors) [1]
*		is=   input format ('b' byte 'X.Yf' ascii)             [8f]
*		to#   temperature offset (added to sensor reading)     [0]
*		ct=   conversion, 'n' none, 'f' to F, 'c' to C         [n]
*		fs=   output (printf) format string                    [%f\n]
*
*	Example rc entry:
*	ayres|outside|outside ayres hall:\
*		dv=/dev/tty00:br#9600:pa=none:\
*		to#-128:ns#3:\
*		fs=Temperature outside Ayres Hall (Knoxville TN) is %.0f degrees F\n:
*
*	Build options:
*
*	Define SYSVSTR if your machine uses strchr() instead of index().
*
*	Define SYSVBFUNC if it uses memcpy() instead of bcopy().
*
*	Define RCFILE to change the path of thermcap.
*
*	If ALLOW_SOCKET_CONNECTIONS is defined when therm is built, then
*	you can use "dv=hostname@portnumber" to connect to a serial port
*	serviced by a network server.  for example, "dv=annex@6005" will
*	connect to port 5 on a Xylogics Annex terminal server.  (the port
*	must be configured as a raw 1200 baud non-flow-controlled port, of
*	course.
*
*	13 jan 1995  Robert Manchek  manchek@cs.utk.edu
*
*	15 mar 96 updated to deal with ascii sensor outputs and C (is=, ct=)
*/

/*
*  Copyright (C) 1995, 1996, 1997 by Robert Manchek.  manchek@CS.UTK.EDU.
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: therm.c,v $
*	Revision 1.6  2000/09/11 04:41:49  ghankins
*	*** empty log message ***
*
*	Revision 1.5  2000/09/11 03:31:52  ghankins
*	*** empty log message ***
*
*	Revision 1.4  2000/09/06 03:27:38  ghankins
*	added numbers to each ERR to see where it breaks
*
*	Revision 1.3  2000/09/05 03:13:31  ghankins
*	*** empty log message ***
*
*	Revision 1.2  2000/09/05 02:34:37  ghankins
*	*** empty log message ***
*
*	Revision 1.1  2000/09/04 16:05:17  ghankins
*	Initial revision
*
*	Revision 1.6  1997/12/15 20:54:24  manchek
*	brought TIMEOUT out to a constant
*
*	Revision 1.5  1997/03/07 02:55:08  manchek
*	added anytoi function, thermcap numerics can now be hex or octal
*
*	Revision 1.4  1997/03/07 02:35:02  manchek
*	added comments about compile options
*
*	Revision 1.3  1997/03/02 03:38:21  manchek
*	added ALLOW_SOCKET_CONNECTIONS contributed code to read remote serial
*	ports (e.g. annex boxes).
*
*	Revision 1.2  1997/01/14 04:42:11  manchek
*	converted to# value to float
*
*	Revision 1.1  1996/11/12 06:27:01  manchek
*	Initial revision
*
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#ifdef	SYSVSTR
#include <string.h>
#define	CINDEX(s,c)	strchr(s,c)
#else
#include <strings.h>
#define	CINDEX(s,c)	index(s,c)
#endif
#ifdef SYSVBFUNC
#include <memory.h>
#define BZERO(d,n)      memset(d,0,n)
#define BCMP(s,d,n)     memcmp(d,s,n)
#define BCOPY(s,d,n)    memcpy(d,s,n)
#else
#define BZERO(d,n)      bzero(d,n)
#define BCMP(s,d,n)     bcmp(s,d,n)
#define BCOPY(s,d,n)    bcopy(s,d,n)
#endif
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/termios.h>
#ifdef ALLOW_SOCKET_CONNECTIONS
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

static char rcsid[] = "$Id: therm.c,v 1.6 2000/09/11 04:41:49 ghankins Exp $";

#define	TRIES	3			/* number of tries to read reply from device */
#define	TIMEOUT	1500000		/* usec to wait for a reply */

#ifndef	RCFILE
#define	RCFILE	"/usr/local/etc/thermcap"
#endif

extern int errno;

int
anytoi(p)
	char *p;
{
	int n = 0;
	int sign = 1;
	int base;
	int d;
	char c;

	while (*p && isspace(*p))
		p++;
	if (*p == '+')
		p++;
	else if (*p == '-') {
		sign = -1;
		p++;
	}
	if (*p == '0') {
		p++;
		if (*p == 'x' || *p == 'X') {
			p++;
			base = 16;
		} else
			base = 8;
	} else
		base = 10;

	while (isxdigit(c = *p++)
	&& (d = c - (isdigit(c) ? '0' : (isupper(c) ? 'A' : 'a') - 10)) < base)
		n = n * base + d;
	return n * sign;
}


/*	breakwords()
*
*	Parse a string into words separated by whitespace.
*	Max number of words is original value of *acp.
*
*	Trashes out the original string.
*	Returns 0 with av[0]..av[*acp - 1] pointing to the words.
*	Returns 1 if too many words.
*	Returns -1 if unbalanced quote.
*/

int
breakwords(s, acp, av, sep)
	char *s;			/* the string to parse */
	int *acp;			/* max num words in, num words found out */
	char **av;			/* pointers to words */
	char *sep;			/* list of separator characters */
{
	int ac = 0;			/* number of words found */
	char *p = s;		/* input scanner */
	char *q;			/* output */
	int n = *acp;		/* max number of words allowed */
	int mode = 0;		/* quote mode */

	while (*p) {
		if (ac >= n) {
			*acp = ac;
			return 1;
		}
		q = p;
		av[ac++] = p;
		while (*p) {
			if (mode) {
				mode = 0;
				switch (*p) {

				case 'n':
					*q++ = '\n';
					break;

				case 'r':
					*q++ = '\r';
					break;

				case 't':
					*q++ = '\t';
					break;

				default:
					*q++ = *p;
					break;
				}

			} else {
				if (CINDEX(sep, *p))
					break;

				if (*p == '\\')
					mode = *p;
				else
					*q++ = *p;
			}
			p++;
		}
		if (*p)
			p++;
		*q = 0;
		if (mode)
			return -1;
	}
	*acp = ac;
	return 0;
}


/*	gettherment()
*
*	Get a thermometer's entry from the thermcap file.
*	Returns a null-terminated list of string attributes, one for each
*	field in the entry.
*/

char **
gettherment(name)
	char *name;
{
	static char *av[20];
	static char buf[4096];

	FILE *ff;
	int l;
	int c, oc;
	int ac;
	int na;
	char *nm[20];

	if (!(ff = fopen(RCFILE, "r")))
		return (char **)0;

	while (1) {

	/* read a device entry from the cap file */

		l = 0;
		oc = ':';
		do {
			if ((c = getc(ff)) == EOF)
				goto notfound;
			while (c == '\\') {
				if ((c = getc(ff)) != '\n') {
					buf[l++] = '\\';
					break;
				}
				c = getc(ff);
			}
			if (oc == ':' && (c == ' ' || c == '\t'))
				continue;
			if (c == '\n')
				c = 0;
			buf[l++] = c;
			oc = c;
		} while (c);

/*
		fprintf(stderr, "<%s>\n", buf);
*/

		ac = sizeof(av)/sizeof(av[0])-1;
		if (breakwords(buf, &ac, av, ":") || ac < 1)
			continue;
		av[ac] = 0;

/*
		for (l = 0; l < ac; l++)
			fprintf(stderr, "[%2d] <%s>\n", l, av[l]);
*/

		na = sizeof(nm)/sizeof(nm[0]);
		if (breakwords(av[0], &na, nm, "|") || na < 1)
			continue;
		if (na > 1)
			na--;

/*
		for (l = 0; l < ac; l++)
			fprintf(stderr, "[%2d] <%s>\n", l, av[l]);
*/

		while (--na >= 0)
			if (!strcmp(nm[na], name))
				break;
		if (na < 0)
			continue;

		fclose(ff);
		return av + 1;
	}
notfound:
	fclose(ff);
	return (char **)0;
}


/*	getabyte()
*
*	Send a byte to the thermometer device and get a response back.
*	Flush any old characters before starting.
*	Time out in less than a second so we don't hang if there's nothing
*	there.
*/

int
getabyte(d, ad, len, resp)
	int d;			/* file */
	int ad;			/* address byte to send */
	int len;		/* length of response */
	char *resp;		/* return response */
{
	fd_set fds;
	struct timeval to;
	int t;
	int cc;
	int got;

	for (t = TRIES; t-- > 0; ) {

		/*
		* flush any junk waiting
		*/
		got = 1000;
		do {
			FD_ZERO(&fds);
			FD_SET(d, &fds);
			to.tv_sec = 0;
			to.tv_usec = 0;
			if ((cc = select(d + 1, &fds, (fd_set *)0, (fd_set *)0, &to)) == 1)
				read(d, resp, 1);
		} while ((got-- > 0 && cc == 1) || (cc == -1 && errno == EINTR));

		/*
		* send address byte
		*/
		FD_ZERO(&fds);
		FD_SET(d, &fds);
		to.tv_sec = 0;
		to.tv_usec = 0;
		if ((cc = select(d + 1, (fd_set *)0, &fds, (fd_set *)0, &to)) != 1) {
			/*if (cc == -1)
				perror("select");
			else
				fprintf(stderr, "sensor port blocked\n");*/
			return -1;
		}
		resp[0] = ad;
		(void)write(d, resp, 1);

		/*
		* read len bytes into resp
		*/
		for (got = 0; got < len; ) {
/*
			fprintf(stderr, "len %d got %d\n", len, got);
*/
			FD_ZERO(&fds);
			FD_SET(d, &fds);
			to.tv_sec = TIMEOUT/1000000;
			to.tv_usec = TIMEOUT%1000000;
			cc = select(d + 1, &fds, (fd_set *)0, (fd_set *)0, &to);
			if (cc == -1)
				if (errno == EINTR)
					continue;
				else {
					perror("select");
					return -1;
				}
			if (cc == 1) {
				cc = read(d, resp + got, len - got);
				if (cc == -1) {
					perror("read");
					return -1;
				}
				if (cc == 0) {
					/*fprintf(stderr, "read EOF\n");*/
					return -1;
				}
				got += cc;

			} else
				break;
		}
/*
		fprintf(stderr, "len %d got %d\n", len, got);
*/
		if (got == len)
			return 0;
	}
	return -1;
}


/*	searchent()
*
*	Search for a field in a thermcap entry and return its value.
*/

char *
searchent(te, name)
	char **te;
	char *name;
{
	int l = strlen(name);

	while (*te)
		if (!strncmp(*te, name, l))
			return *te + l;
		else
			te++;
	return (char *)0;
}


struct baudval {
	char *name;
	int speed;
	int setting;
};

struct baudval rates[] = {
	{ "50", 50, B50 },
	{ "75", 75, B75 },
	{ "110", 110, B110 },
	{ "134", 134, B134 },
	{ "150", 150, B150 },
	{ "200", 200, B200 },
	{ "300", 300, B300 },
	{ "600", 600, B600 },
	{ "1200", 1200, B1200 },
	{ "2400", 2400, B2400 },
	{ "4800", 4800, B4800 },
	{ "9600", 9600, B9600 },
#ifdef B19200
	{ "19200", 19200, B19200 },
#endif
#ifdef B38400
	{ "38400", 38400, B38400 },
#endif
#ifdef EXTA
	{ "exta", -1, EXTA },
#endif
#ifdef EXTB
	{ "extb", -1, EXTB },
#endif
	{ 0, 0 }
};


/*	setupline()
*
*	Configure the therm device tty.
*/

int
setupline(d, te)
	int d;
	char **te;
{
	struct termios tio;
	int i;
	char *br;
	char *pa;
	struct baudval *bp;

	if (tcgetattr(d, &tio)) {
		if (errno != ENOTTY && errno != EINVAL)
			perror("tcgetattr");
		return 1;
	}

/*
	printtio(&tio);
*/

	tio.c_iflag = IGNPAR|INPCK;
/*
	tio.c_oflag = ONL0 | CR0 | TAB0 | BS0 | VT0 | FF0;
*/
	tio.c_oflag = 0;
#ifdef	CBAUD
	tio.c_cflag = (tio.c_cflag & CBAUD) | CREAD | CLOCAL | CS8;
#else
	tio.c_cflag = CREAD | CLOCAL | CS8;
#endif
	tio.c_lflag = 0;
	tio.c_cc[VMIN] = 0;
	tio.c_cc[VTIME] = 0;

	if ((pa = searchent(te, "pa="))) {
		switch (*pa) {

		case 'e':
			tio.c_cflag |= PARENB;
			break;

		case 'o':
			tio.c_cflag |= (PARODD|PARENB);
			break;

/*
		case 'm':
			tio.c_cflag |= (PARODD|PAREXT|PARENB);
			break;

		case 's':
			tio.c_cflag |= (PAREXT|PARENB);
			break;
*/

		default:
			break;
		}
	}

	if ((br = searchent(te, "br#"))) {
		i = anytoi(br);
		for (bp = rates; bp->name; bp++)
			if (bp->speed == i)
				break;
		if (bp->name) {
			cfsetospeed(&tio, (speed_t)bp->setting);
			cfsetispeed(&tio, (speed_t)bp->setting);
		} /* else {
			fprintf(stderr, "unknown baud rate: %d\n", i);
		} */
	}

	if (tcsetattr(d, TCSAFLUSH, &tio))
		perror("tcsetattr");

	if (tcgetattr(d, &tio))
		perror("tcgetattr");

/*
	printtio(&tio);
*/
	return 0;
}


#ifdef ALLOW_SOCKET_CONNECTIONS
int
opensocketto(dev)
	char *dev;
{
	struct hostent *hp;
	struct sockaddr_in sa;
	unsigned long addr;
	char *ptr;
	int d;

	BZERO((char *)&sa, sizeof(sa));
	ptr = CINDEX(dev, '@');
	*ptr++ = '\0';
	if ((addr = inet_addr(dev)) == ~0) {
		if ((hp = gethostbyname(dev)) != NULL) {
			BCOPY(hp->h_addr_list[0], (char *)&addr, sizeof(addr));
			endhostent();

		} else {
			/*fprintf(stderr, "unknown host `%s'\n", dev);*/
			endhostent();
			return -1;
		}
	}
	sa.sin_family = AF_INET;
	sa.sin_port = htons((short)(*ptr ? anytoi(ptr) : 0));
	sa.sin_addr.s_addr = addr;

	*--ptr = '@';

	if ((d = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		/*fprintf(stderr, "error opening socket (%s)\n",
			strerror(errno));*/
		return -1;
	}
	if (connect(d, (struct sockaddr *) & sa, sizeof(sa)) == -1) {
		/*fprintf(stderr, "error connecting socket (%s)\n",
			strerror(errno));*/
		return -1;
	}
	return d;
}
#endif /*ALLOW_SOCKET_CONNECTIONS*/

/*	openline()
*
*	Open the thermometer device tty and configure it.
*/

int
openline(te)
	char **te;
{
	char *dev;
	int d;

	if (!(dev = searchent(te, "dv="))) {
		/*fprintf(stderr, "bad configuration\n");*/
		return -1;
	}

#ifdef ALLOW_SOCKET_CONNECTIONS
	/*
	 * Look for a host@port entry
	 */
	if (CINDEX(dev, '@') != NULL)
		return opensocketto(dev);
#endif

	if ((d = open(dev, O_RDWR, 0)) == -1) {
		perror(dev);
		return -1;
	}

	setupline(d, te);

	return d;
}


static int
printtio(tiop)
	struct termios *tiop;
{
	int i;
	struct baudval *bp;

	printf("c_iflag 0x%x, c_oflag 0x%x, c_cflag 0x%x, c_lflag 0x%x\n", 
			tiop->c_iflag,
			tiop->c_oflag,
			tiop->c_cflag,
			tiop->c_lflag);

	i = cfgetispeed(tiop);
	for (bp = rates; bp->name; bp++)
		if (bp->setting == i)
			break;
	if (bp->name)
		printf(" input speed %s", bp->name);
	i = cfgetospeed(tiop);
	for (bp = rates; bp->name; bp++)
		if (bp->setting == i)
			break;
	if (bp->name)
		printf(" output speed %s\n", bp->name);
	printf(" c_cc[");
	for (i = 0; i < NCCS; i++)
		printf(" %d", (int)(tiop->c_cc[i]));
	printf(" ]\n");
	return 0;
}


/*	readsensor()
*
*	Get a reading from the sensor.
*/

int
readsensor(ifmt, fd, ad, r)
	char *ifmt;				/* sensor output format */
	int fd;					/* file */
	int ad;					/* address of sensor */
	double *r;				/* reading return */
{
	int w1 = 0;			/* field width */
	int w2 = 0;			/* precision */
	char buf[40];

	w1 = 1;
	if (isdigit(*ifmt)) {
		w1 = atoi(ifmt);
		while (isdigit(*++ifmt)) ;
	}
	w2 = 0;
	if (*ifmt == '.') {
		ifmt++;
		if (isdigit(*ifmt)) {
			w2 = atoi(ifmt);
			while (isdigit(*++ifmt)) ;
		}
	}
	switch (*ifmt) {
	case 'b':
		if (getabyte(fd, ad, 1, buf) < 0)
			goto noresp;
/*
		fprintf(stderr, "0x%x\n", buf[0] & 0xff);
*/
		*r = buf[0] & 0xff;
		break;

	case 'f':
		if (getabyte(fd, ad, w1, buf) < 0)
			goto noresp;
		buf[w1] = 0;
		if (sscanf(buf, "%lf", r) != 1) {
			/*fprintf(stderr, "no sensor\n");*/
			*r = 0;
			return -1;
		}
		break;

	default:
		/*fprintf(stderr, "unknown sensor format: '%c'\n", *ifmt);*/
		return -1;
	}
	return 0;

noresp:
	/*fprintf(stderr, "no response from device\n");*/
	return -1;
}


#ifdef	MEDIAN
static int
doublecmp(a, b)
	double *a, *b;
{
	if (*a > *b)
		return 1;
	else if (*a < *b)
		return -1;
	return 0;
}
#endif


#ifdef NOT_NEEDED_ANYMORE
char *
therm_main(therm_argc, therm_argv)
	int therm_argc;
	char **therm_argv;
{
	char *ifmt = 0;				/* sensor input format */
	char *conv = 0;				/* type of conversion */
	char *therm = 0;			/* name of therm to read */
	char *format = 0;			/* output format string */
	double tof = 0;				/* temperature offset */
	int ns = 1;					/* number of samples */
	int ad = 0;					/* address byte */
	int i;
	int d;
	char **te;
	char *p;
	double t;
	double ts;
	double *ta;
	int j;
        char str[11];


	for (j = 1; j < therm_argc; j++) {
		ifmt = "8f";
		conv = "n";
		format = "%f\n";
		tof = 0;
		ns = 1;
		ad = 0;
		therm = therm_argv[j];

		if (!(te = gettherment(therm))) {
			fprintf(stderr, "unknown device: %s\n", therm);
                        return("E1");
			continue;
		}
		if ((p = searchent(te, "is=")))
			ifmt = p;
		if ((p = searchent(te, "ct=")))
			conv = p;
		if ((p = searchent(te, "fs=")))
			format = p;
		if ((p = searchent(te, "to#")))
			tof = atof(p);
		if ((p = searchent(te, "ns#"))) {
			ns = anytoi(p);
			if (ns < 1)
				ns = 1;
			else if (ns > 1000)
				ns = 1000;
		}

		if ((p = searchent(te, "ad#")))
			ad = anytoi(p);

		ta = (double *)malloc(ns * sizeof(double));

		if ((d = openline(te)) == -1)
			continue;

		ts = 0;
		for (i = ns; i-- > 0; ) {
			if (readsensor(ifmt, d, ad, &t) < 0) {
		            (void)close(d);
                            return("E2");
                        }
			t += tof;
/*
			fprintf(stderr, "%f F\n", t);
*/
			ta[i] = t;
			ts += t;
		}

#ifdef	MEDIAN
		qsort((char *)ta, ns, sizeof(double), doublecmp);
		ts = ta[ns / 2];
#else
		ts /= ns;
#endif

		switch (conv[0]) {
		case 'f':
			ts = (ts * 9) / 5 + 32;
			break;
		case 'c':
			ts = ((ts - 32) * 5) / 9;
			break;
		default:
			break;
		}
		/*printf(format, ts);*/
                snprintf(str, 10, "%2.2f%c", ts, toupper(conv[0]));
                return(str);
	}
	exit(0);
}
#endif
